#ifdef WITH_MPI
#include <mpi.h>
#endif
#include <omp.h>
#include <functional>
#include <algorithm>
#include <numeric>
#include <unistd.h>
#include "defs.h"
#include "dbg_helper.h"
#include "check.inl"

extern void peer2peer_exchange(const std::vector<int> &params, olocal &output);
extern void hip_memcopy(const std::vector<int> &params, olocal &output);
extern int hip_stream(const std::vector<int> &params, olocal &output);
extern void core_affinity(const std::vector<int> &params, olocal &output);
extern void nic_affinity(const std::vector<int> &params, olocal &output);
extern void device_affinity(const std::vector<int> &params, olocal &output);

void do_test(std::function<void(const std::vector<int> &, olocal &output)> fn, const std::vector<int> &params, olocal &output) {
#pragma omp parallel 
    {    
        for (int j = 0; j < 1; j++) {
#pragma omp master
            {
                fn(params, output);
            }
        }
    }
#ifdef WITH_MPI
    MPI_CALL(MPI_Barrier(MPI_COMM_WORLD));
#endif
}

void print_final_output(int rank, const std::vector<oglobal> &output) {
    for (const auto &og : output) {
        std::stringstream ss_outer;
        ss_outer << og.header << ": {";
        std::stringstream ss_inner;
        ss_inner << std::setprecision(1) << std::fixed;
        for (const auto value : og.values) {
            ss_inner << "," << value;
        }
        ss_outer << ss_inner.str().erase(0, 1) << "}";
        print(rank, ss_outer.str());
    }
}

void local2output(olocal &output_local, std::vector<oglobal> &output) {
    if (output_local.values.empty())
        return;
    std::vector<double> aggregator;
    for (const auto &v : output_local.values) {
        for (const auto value : v) {        
            aggregator.push_back(value);
        }
    }
    output.push_back({output_local.header, aggregator});
    output_local.values.resize(0);
}

void init(int &argc, char **&argv, int &worldsize, int &rank) {
    (void)argc;
    (void)argv;
    (void)worldsize;
#ifdef WITH_MPI
    MPI_CALL(MPI_Init(&argc, &argv));
#endif
//#ifdef WITH_TIMEOUT
//    alarm(WITH_TIMEOUT); // NOTE: this alarm is to interrupt test in case of deadlock
//#endif
#ifdef WITH_MPI
    MPI_CALL(MPI_Comm_size(MPI_COMM_WORLD, &worldsize));
    MPI_CALL(MPI_Comm_rank(MPI_COMM_WORLD, &rank));
#endif
#ifdef WITH_OUTPUT
    int n = omp_get_max_threads();
    std::cout << ">> rank=" << rank << " worldsize: "<< worldsize << " omp nthreads=" << n << std::endl;
#endif
#ifdef WITH_MPI
    MPI_CALL(MPI_Barrier(MPI_COMM_WORLD));
#endif
}

int main(int argc, char **argv)
{
    int worldsize = 0, rank = 0;
    // NOTE: this alarm is to interrupt test in case of deadlock
    // We also install signal handlers and have a stacktrace code there
    dbg_helper dbg;
    try {
        init(argc, argv, worldsize, rank);

        // NOTE: this alarm is to interrupt test in case of deadlock
        // We also install signal handlers and have a stacktrace code there
#ifdef WITH_TIMEOUT
        dbg.start(dbg_helper::SIGHANDLERS | dbg_helper::TIMER, WITH_TIMEOUT);
#else
        dbg.start(dbg_helper::SIGHANDLERS);
#endif

        std::vector<oglobal> output;
        olocal output_local;
        do_test(core_affinity, {rank}, output_local);
        //do_test(nic_affinity, {rank}, output_local);
#if WITH_GPU
        do_test(device_affinity, {rank}, output_local);
        do_test(hip_stream, {rank}, output_local);
#endif
        local2output(output_local, output);
#if WITH_GPU
        for (const int use_pinned_memory : {1,0}) { 
            for (int interleaving = 1; interleaving < NRANKS_PER_NODE * 2; interleaving *= 2) {
                do_test(hip_memcopy, {rank, interleaving, NRANKS_PER_NODE, use_pinned_memory}, output_local);
            }
            local2output(output_local, output);
        }
#endif
#ifdef WITH_MPI
        for (const int bidirectional : {0,1}) { 
#if WITH_GPU
            for (const int use_gpu_memory : {1,0}) { 
#else
            for (const int use_gpu_memory : {0}) { 
#endif
                std::string kind = (use_gpu_memory ? "Device2Device" : "Host2Host");
                print(rank, kind + " MPI exchange, intra-node:");
                for (int intermix = 0; intermix < NRANKS_PER_NODE/2; intermix++) {
                    do_test(peer2peer_exchange, 
                            {rank, NRANKS_PER_NODE, intermix, NRANKS_PER_NODE, 
                             use_gpu_memory, bidirectional}, output_local);
                }
                local2output(output_local, output);
                if (worldsize > NRANKS_PER_NODE) {
                    print(rank, kind + "MPI exchange, cross-node:");
                    for (int intermix = 0; intermix < NRANKS_PER_NODE; intermix += 2) {
                        do_test(peer2peer_exchange, 
                                {rank, NRANKS_PER_NODE*2, intermix, NRANKS_PER_NODE, 
                                 use_gpu_memory, bidirectional}, output_local);
                    }
                    local2output(output_local, output);
                }
            }
        }
#endif
        print_final_output(rank, output);
#ifdef WITH_CHECKS
        std::vector<std::string> exclude;
#ifdef WITH_MPI
        exclude.push_back("MPI_.*");
#else
        if (worldsize <= NRANKS_PER_NODE) {
            exclude.push_back("MPI_.*_crossnode_.*");
        }
#endif
   
#if !WITH_GPU
        exclude.push_back(".*_GPU_.*");
        exclude.push_back("GPU_.*");
        exclude.push_back(".*_D2H_.*");
#endif
     
        bool result = do_all_checks(rank, output, exclude);
        int mpiresult = (result?1:0);
#ifdef WITH_MPI
        MPI_CALL(MPI_Allreduce(MPI_IN_PLACE, &mpiresult, 1, MPI_INT, MPI_LAND, MPI_COMM_WORLD));
        MPI_CALL(MPI_Finalize());
#endif
        if (!mpiresult) {
            printx(rank, "CHECK FAILED ON ONE OF RANKS");
        }
        return !mpiresult;
#endif
    }
    catch (std::exception &ex) {
        printx(rank, std::string("EXCEPTION: ") + ex.what()); 
        return 1;
    }
    return 0;
}
